package automaton;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

public class DeterministicAutomaton {

	private State initialState = null;

	/*
	 * In the map transitions, at each state s we associate a map m where the
	 * values are the transitions having s as source and the corresponding key
	 * the labels of the transitions.
	 * 
	 */
	private final Map<State, Map<Object, Transition>> transitions;

	public DeterministicAutomaton(Transition... transitions)
			throws NotDeterministicTransitionException,
			NotDeterministicInitialStateException, UnknownInitialStateException {
		this.transitions = new HashMap<State, Map<Object, Transition>>();
		for (Transition t : transitions) {
			addState(t.source());
			addState(t.target());
			Map<Object, Transition> map = this.transitions.get(t.source());
			if (map.containsKey(t.label())) {
				throw new NotDeterministicTransitionException(t, map.get(t
						.label()));
			} else {
				map.put(t.label(), t);
			}
		}
		if (initialState == null) {
			throw new UnknownInitialStateException();
		}
	}

	protected final void addState(State s)
			throws NotDeterministicInitialStateException {
		if (!transitions.containsKey(s)) {
			if (s.initial()) {
				if (initialState == null) {
					initialState = s;
				} else {
					throw new NotDeterministicInitialStateException(s,
							initialState);
				}
			}
			transitions.put(s, new HashMap<Object, Transition>());
		}
	}

	public State initialState() {
		return initialState;
	}

	/**
	 * @exception NoSuchElementException
	 *                if the <tt>source</tt> does not belong to the automaton.
	 */
	public Transition transition(State source, Object label) {
		if (!transitions.containsKey(source)) {
			throw new NoSuchElementException();
		}
		return transitions.get(source).get(label);
	}

	public boolean recognize(Object... word) {
		return recognize(Arrays.asList(word).iterator());
	}

	public boolean recognize(Iterator<Object> word) {
		State s = initialState;
		while (word.hasNext()) {
			Transition t = transition(s, word.next());
			if (t == null) {
				return false;
			} else {
				s = changeCurrentState(t);
			}
		}
		return s.terminal();
	}
	
	protected State changeCurrentState(Transition t) {
		return t.target();
	}
}
