package regexp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import automaton.NotDeterministicAutomaton;
import automaton.State;
import automaton.Transition;
import automaton.TransitionImpl;
import automaton.UnknownInitialStateException;

public class AutomatonBuilder {

	private static final Operator OR = new Operator() {

		public int arity() {
			return 2;
		}

		public AutomatonData build(AutomatonData[] args) {
			AutomatonData ad1 = args[0];
			AutomatonData ad2 = args[1];
			ad1.initialStates.addAll(ad2.initialStates);
			ad1.terminalStates.addAll(ad2.terminalStates);
			ad1.transitions.addAll(ad2.transitions);
			return ad1;
		}

		public String symbol() {
			return "|";
		}

	};

	private static final Operator DOT = new Operator() {

		public int arity() {
			return 2;
		}

		public AutomatonData build(AutomatonData[] args) {
			AutomatonData ad1 = args[0];
			AutomatonData ad2 = args[1];
			ad1.transitions.addAll(ad2.transitions);
			for (State s1 : ad1.terminalStates)
				for (State s2 : ad2.initialStates)
					ad1.transitions.add(new TransitionImpl<String>(s1, s2, ""));
			ad1.terminalStates = ad2.terminalStates;
			return ad1;
		}

		public String symbol() {
			return ".";
		}

	};

	private static final Operator ITER = new Operator() {

		public int arity() {
			return 1;
		}

		public AutomatonData build(AutomatonData[] args) {
			AutomatonData ad = args[0];
			StateImpl s = new StateImpl();
			for (State ts : ad.terminalStates)
				ad.transitions.add(new TransitionImpl<String>(ts, s, ""));
			for (State is : ad.initialStates)
				ad.transitions.add(new TransitionImpl<String>(s, is, ""));
			ad.initialStates.clear();
			ad.initialStates.add(s);
			ad.terminalStates.clear();
			ad.terminalStates.add(s);
			return ad;
		}

		public String symbol() {
			return "*";
		}

	};

	private static final Operator[] ops = { DOT, OR, ITER };

	private static final Map<String, Operator> operators = new HashMap<String, Operator>();

	static {
		for (Operator op : ops)
			operators.put(op.symbol(), op);
	}

	public static NotDeterministicAutomaton<String> createAutomaton(
			String expression) throws SyntaxErrorException,
			UnknownInitialStateException {
		Iterator<String> its = Arrays.asList(expression.split("\\s"))
				.iterator();
		AutomatonData ad = evaluate(its);
		if (its.hasNext())
			throw new SyntaxErrorException(" : " + its.next());
		for (StateImpl s : ad.initialStates) {
			s.setInitial();
		}
		for (StateImpl s : ad.terminalStates) {
			s.setTerminal();
		}
		return new NotDeterministicAutomaton<String>("", ad.transitions
						.toArray(new Transition[ad.transitions.size()]));
	}

	private static AutomatonData evaluate(Iterator<String> it)
			throws SyntaxErrorException {
		if (!it.hasNext()) {
			throw new SyntaxErrorException();
		}
		String s = it.next();
		if (operators.containsKey(s)) {
			Operator op = operators.get(s);
			AutomatonData[] args = new AutomatonData[op.arity()];
			for (int j = 0; j < op.arity(); j++) {
				args[j] = evaluate(it);
			}
			return op.build(args);
		} else {
			AutomatonData ad = new AutomatonData();
			StateImpl is = new StateImpl();
			StateImpl ts = new StateImpl();
			ad.initialStates = new HashSet<StateImpl>();
			ad.initialStates.add(is);
			ad.terminalStates = new HashSet<StateImpl>();
			ad.terminalStates.add(ts);
			ad.transitions = new ArrayList<Transition<String>>();
			ad.transitions.add(new TransitionImpl<String>(is, ts, s));
			return ad;
		}
	}

}
